(*  title     : A Shallow encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory Example
imports
   FOTheorems
begin

section \<open>Example\<close>

 sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 e z3 vampire spass  " ]

declare [[show_types]]

(*
declare [[show_consts]]
*)

locale Test =
 fixes v :: "state \<Rightarrow> nat"
 fixes v1 :: "state \<Rightarrow> nat"
 fixes y :: "state \<Rightarrow> bool" 
 fixes z :: "state \<Rightarrow> int"
 fixes F2 :: "nat statefun \<Rightarrow> temporal "
 fixes F3 :: "bool statefun \<Rightarrow> temporal " 
 fixes F4 :: "int statefun \<Rightarrow> temporal" 
 fixes F5 :: "nat statefun \<Rightarrow> temporal "
 fixes Init2 :: "nat statefun \<Rightarrow> temporal "
 fixes Init3 :: "bool statefun \<Rightarrow> temporal " 
 defines "F2 \<equiv> (\<lambda> v. TEMP \<box> ( #0 \<le> $v )) "
 defines "F3 \<equiv> (\<lambda> p. TEMP \<box> ( $p \<or> \<not> $p))" 
 defines "F4 \<equiv> (\<lambda> z. TEMP \<box> ( #0\<le> $z \<or> $z < #0))"  
 defines "F5 \<equiv> (\<lambda> v. TEMP $v=#0 \<and> v gets $v+#1)"
 defines "Init2 \<equiv> (\<lambda> v. TEMP $v = #0)"
 defines "Init3 \<equiv> (\<lambda> p. TEMP $p)"


locale Test1 =
 fixes v :: "state \<Rightarrow> nat"
 fixes F5 :: "nat statefun \<Rightarrow> nat \<Rightarrow> temporal "
 defines "F5 \<equiv> (\<lambda> v n. TEMP $v=#0 \<and> v gets $v+#1 \<and> fin($v=#n))"

definition F1 :: "nat statefun \<Rightarrow> temporal "
where " F1 w \<equiv> TEMP \<box> ( #0 \<le> $w )" 

definition Init1 :: "nat statefun \<Rightarrow> temporal "
where "Init1 w \<equiv> TEMP $w = #0"

lemma (in Test) currentval_test :
 " (s \<Turnstile> ($v = #0)) = ( (v (nth s 0)) = 0)"
by (simp add: current_val_d_def)

lemma (in Test) nextempty_test :
 " (\<langle>s0\<rangle> \<Turnstile> v$) =  (\<some> x. x=x)"
by (simp add: next_val_d_def)

lemma (in Test) nextempty_test_1 :
 " (\<langle>s0\<rangle> \<Turnstile> v$ = v$)"
by simp

lemma (in Test) nextempty_test_2 :
 " (\<langle>s0\<rangle> \<Turnstile> v$ = v1$)"
by (simp add: Test.nextempty_test)

lemma (in Test) nextcurrent_test:
 " (\<langle>s0,s1\<rangle> \<Turnstile> skip \<and> ($v=#0) \<and> (v$=$v+#1)) = (((v s0) = 0) \<and> ((v s1) = 1  ))  "
unfolding current_val_d_def next_val_d_def skip_defs by auto

lemma (in Test) nextcurrentfinpenult_test:
 " (\<langle>s0,s1,s2,s3\<rangle> \<Turnstile> len(3) \<and> v =: !v-#1 \<and> v \<leftarrow> #3 \<and> $v=#0 \<and> v := $v+#1 ) = 
     (((v s0) = 0) \<and> ((v s1) = 1 \<and> (v (s2)) = 2 \<and> ((v s3) = 3   )))  "
unfolding current_val_d_def next_val_d_def fin_val_d_def penult_val_d_def 
          assign_d_def prev_assign_d_def temporal_assign_d_def len_defs by auto


lemma (in Test) stable_test:
 "(\<langle>s0,s1,s2,s3\<rangle> \<Turnstile> len(3) \<and> stable v \<and> $v=#0) = 
   ((v s0) = 0 \<and> (v s1) = 0 \<and> (v s2) = 0 \<and> (v s3) = 0)"
by (auto simp: stable_defs len_defs  
           current_val_d_def next_val_d_def Nitpick.case_nat_unfold)

lemma (in Test) revnextcurrentfinpenult_test:
 " (\<langle>s0,s1,s2,s3\<rangle> \<Turnstile> (len 3 \<and> v! = !v-#1 \<and> !v = #3 \<and> $v=#0 \<and> v$=$v+#1)\<^sup>r) = 
     (((v s3) = 0) \<and> ((v s2) = 1 \<and> (v (s1)) = 2 \<and> ((v s0) = 3   )))  "
unfolding reverse_d_def len_defs current_val_d_def next_val_d_def 
          penult_val_d_def fin_val_d_def by auto

lemma revnextcurrentfinpenult:
 "\<turnstile> more \<longrightarrow> (v$ = $v)\<^sup>r = (v!=!v)"
proof -
 have 1: "\<turnstile> more \<longrightarrow> (v$ = $v)\<^sup>r = ( (v$)\<^sup>r = ($v)\<^sup>r)" 
      by (metis (mono_tags, lifting) intI inteq_reflection rev_fun2 unl_lift2)
 have 2: "\<turnstile> more \<longrightarrow> ( (v$)\<^sup>r =  (v!) )" by (simp add: rev_next)
 have 3: "\<turnstile> more \<longrightarrow> ( ($v)\<^sup>r =  (!v))" using rev_current by fastforce
 have 4: " \<turnstile> more \<longrightarrow> (( (v$)\<^sup>r = ($v)\<^sup>r) \<longrightarrow> ( (v!) = (!v) ))" using "2" "3" by fastforce
 have 5: " \<turnstile> more \<longrightarrow> (( (v!) = (!v) ) \<longrightarrow>  ( (v$)\<^sup>r = ($v)\<^sup>r)    )" using "2" "3" by fastforce
 have 6: " \<turnstile> more \<longrightarrow> (( (v$)\<^sup>r = ($v)\<^sup>r) = ( (v!) = (!v) ))" using "4" "5" by fastforce
 from 1 6 show ?thesis by fastforce
qed

lemma init1:
 "(\<langle>s0,s1,s2\<rangle> \<Turnstile> len(2) \<and> Init1 w) = ((w s0) = 0) "
by (simp add: Init1_def current_val_d_def len_defs)

lemma  exist_test_F1 :
 "\<turnstile> \<exists>\<exists> w. F1 w"
proof -
 have 1: "\<And> w. \<turnstile> F1 w" by (simp add: always_defs current_val_d_def F1_def Valid_def)
 from 1 show ?thesis using EExI[unlift_rule] by blast
qed
 
lemma (in Test) exist_test_F2 :
 "\<turnstile> \<exists>\<exists> v. F2 v"
proof -
 have 1: "\<turnstile> F2 v" by (simp add: always_defs current_val_d_def F2_def Valid_def)
 from 1 show ?thesis using EExI[unlift_rule] by blast
qed

lemma (in Test) exist_test_F3 :
 "\<turnstile> \<exists>\<exists> y. F3 y"
proof -
 have 1: "\<turnstile> F3 y" by (simp add: always_defs current_val_d_def F3_def Valid_def)
 from 1 show ?thesis using EExI[unlift_rule] by blast
qed

lemma (in Test1) test_E_F5_1:
 "(
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i))) \<and>
           x (Interval.nth w (intlen w)) = n) \<longrightarrow>
   (
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>intlen w. x (Interval.nth w (i)) = i) \<and>
           x (Interval.nth w (intlen w)) = n)"
apply simp
proof 
   assume 0: "x (Interval.nth w (0::nat)) = (0::nat) \<and>
    (\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i))) \<and> x (Interval.nth w (intlen w)) = n "
   have 1: "x (Interval.nth w (0::nat)) = (0::nat) " using 0 by auto
   have 2: "x (Interval.nth w (intlen w)) = n " using 0 by auto
   have 3: "(\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i)))" using 0 by auto
   show "\<forall>i\<le>intlen w. x (Interval.nth w i) = i "
   proof
     fix i
     show "i \<le> intlen w \<longrightarrow> x (Interval.nth w i) = i "
     proof 
      (induct i)
      case 0
      then show ?case using 1 by simp
      next
      case (Suc i)
      then show ?case by (simp add: "3")
     qed
  qed
qed

lemma (in Test1) test_E_F5_2:
 "(
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>intlen w. x (Interval.nth w (i)) = i) \<and>
           x (Interval.nth w (intlen w)) = n) \<longrightarrow> (
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i))) \<and>
           x (Interval.nth w (intlen w)) = n) 
   "
by simp

lemma (in Test1) test_E_F5_3:
 "(
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i))) \<and>
           x (Interval.nth w (intlen w)) = n) =
   (
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>intlen w. x (Interval.nth w (i)) = i) \<and>
           x (Interval.nth w (intlen w)) = n)"
using test_E_F5_1 test_E_F5_2 by auto

lemma (in Test1) test_E_F5_4:
 "(\<exists>x::state \<Rightarrow> nat.
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i))) \<and>
           x (Interval.nth w (intlen w)) = n) =
   (\<exists>x::state \<Rightarrow> nat.
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>intlen w. x (Interval.nth w (i)) = i) \<and>
           x (Interval.nth w (intlen w)) = n)"
by (simp add: Test1.test_E_F5_3)


lemma (in Test1) test_E_F5:
 "\<turnstile> (\<exists>\<exists> v. (F5 v n)) \<longrightarrow> (len n)"
apply (simp add: Valid_def F5_def exist_state_d_def gets_defs current_val_d_def fin_defs sub_def len_defs)
proof
  fix w
  show "(\<exists>x::state \<Rightarrow> nat.
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i))) \<and>
           x (Interval.nth w (intlen w)) = n) \<longrightarrow>
       (intlen w = n) "
  proof -
   have 1: "(\<exists>x::state \<Rightarrow> nat.
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i<intlen w. x (Interval.nth w (Suc i)) = Suc (x (Interval.nth w i))) \<and>
           x (Interval.nth w (intlen w)) = n) =
           (\<exists>x::state \<Rightarrow> nat.
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>intlen w. x (Interval.nth w (i)) = i) \<and>
           x (Interval.nth w (intlen w)) = n) " using test_E_F5_4 by auto
   have 2: "(\<exists>x::state \<Rightarrow> nat.
           x (Interval.nth w (0::nat)) = (0::nat) \<and>
           (\<forall>i\<le>intlen w. x (Interval.nth w (i)) = i) \<and>
           x (Interval.nth w (intlen w)) = n) \<longrightarrow> (intlen w =n)"
           by auto
   from 1 2 show ?thesis by auto
 qed
qed
           

end
